searchNodes=[{"type":"module","title":"CubDB","doc":"CubDB is an embedded key-value database for the Elixir language. It is designed for robustness, and for minimal need of resources. Features Both keys and values can be any Elixir (or Erlang) term. Basic get/3 , put/3 , and delete/2 operations, selection of ranges of entries sorted by key with select/2 . Atomic, Consistent, Isolated, Durable (ACID) transactions. Multi version concurrency control (MVCC) allowing concurrent read operations, that do not block nor are blocked by writes. Unexpected shutdowns or crashes won't corrupt the database or break atomicity of transactions. Manual or automatic compaction to reclaim disk space. To ensure consistency, performance, and robustness to data corruption, CubDB database file uses an append-only, immutable B-tree data structure. Entries are never changed in-place, and read operations are performed on zero cost immutable snapshots. More information can be found in the following sections: Frequently Asked Questions How To Usage Start CubDB by specifying a directory for its database file (if not existing, it will be created): { :ok , db } = CubDB . start_link ( &quot;my/data/directory&quot; ) Alternatively, to specify more options, a keyword list can be passed: { :ok , db } = CubDB . start_link ( data_dir : &quot;my/data/directory&quot; , auto_compact : true ) Important Avoid starting multiple CubDB processes on the same data directory. Only one CubDB process should use a specific data directory at any time. CubDB functions can be called concurrently from different processes, but it is important that only one CubDB process is started on the same data directory. The get/2 , put/3 , and delete/2 functions work as you probably expect: CubDB . put ( db , :foo , &quot;some value&quot; ) #=&gt; :ok CubDB . get ( db , :foo ) #=&gt; &quot;some value&quot; CubDB . delete ( db , :foo ) #=&gt; :ok CubDB . get ( db , :foo ) #=&gt; nil Both keys and values can be any Elixir (or Erlang) term: CubDB . put ( db , { &quot;some&quot; , &#39;tuple&#39; , :key } , %{ foo : &quot;a map value&quot; } ) #=&gt; :ok CubDB . get ( db , { &quot;some&quot; , &#39;tuple&#39; , :key } ) #=&gt; %{foo: &quot;a map value&quot;} Multiple operations can be performed atomically with the transaction/2 function and the CubDB.Tx module: # Swapping `:a` and `:b` atomically: CubDB . transaction ( db , fn tx -&gt; a = CubDB.Tx . get ( tx , :a ) b = CubDB.Tx . get ( tx , :b ) tx = CubDB.Tx . put ( tx , :a , b ) tx = CubDB.Tx . put ( tx , :b , a ) { :commit , tx , :ok } end ) #=&gt; :ok Alternatively, it is possible to use put_multi/2 , delete_multi/2 , and the other [...]_multi functions, which also guarantee atomicity: CubDB . put_multi ( db , [ a : 1 , b : 2 , c : 3 , d : 4 , e : 5 , f : 6 , g : 7 , h : 8 ] ) #=&gt; :ok Range of entries sorted by key are retrieved using select/2 : CubDB . select ( db , min_key : :b , max_key : :e ) |&gt; Enum . to_list ( ) #=&gt; [b: 2, c: 3, d: 4, e: 5] The select/2 function can select entries in normal or reverse order, and returns a lazy stream, so one can use functions in the Stream and Enum modules to map, filter, and transform the result, only fetching from the database the relevant entries: # Take the sum of the last 3 even values: CubDB . select ( db , reverse : true ) # select entries in reverse order |&gt; Stream . map ( fn { _key , value } -&gt; value end ) # discard the key and keep only the value |&gt; Stream . filter ( fn value -&gt; is_integer ( value ) &amp;&amp; Integer . is_even ( value ) end ) # filter only even integers |&gt; Stream . take ( 3 ) # take the first 3 values |&gt; Enum . sum ( ) # sum the values #=&gt; 18 Read-only snapshots are useful when one needs to perform several reads or selects, ensuring isolation from concurrent writes, but without blocking them. When nothing needs to be written, using a snapshot is preferable to using a transaction, because it will not block writes. Snapshots come at no cost: nothing is actually copied or written on disk or in memory, apart from some small internal bookkeeping. After obtaining a snapshot with with_snapshot/2 , one can read from it using the functions in the CubDB.Snapshot module: # the key of y depends on the value of x, so we ensure consistency by getting # both entries from the same snapshot, isolating from the effects of concurrent # writes { x , y } = CubDB . with_snapshot ( db , fn snap -&gt; x = CubDB.Snapshot . get ( snap , :x ) y = CubDB.Snapshot . get ( snap , x ) { x , y } end ) The functions that read multiple entries like get_multi/2 , select/2 , etc. are internally using a snapshot, so they always ensure consistency and isolation from concurrent writes, implementing multi version concurrency control (MVCC). Because CubDB uses an immutable, append-only data structure, write operations cause the data file to grow. When necessary, CubDB runs a compaction operation to optimize the file size and reclaim disk space. Compaction runs in the background, without blocking other operations. By default, CubDB runs compaction automatically when necessary (see documentation of set_auto_compact/2 for details). Alternatively, it can be started manually by calling compact/1 .","ref":"CubDB.html"},{"type":"function","title":"CubDB.back_up/2","doc":"Creates a backup of the database into the target directory path The directory is created upon calling back_up/2 , and an error tuple is returned if it already exists. The function will block until the backup is completed, then return :ok. The backup does not block other readers or writers, and reflects the database state at the time it was started, without any later write. After the backup completes successfully, it is possible to open it by starting a CubDB process using the target path as its data directory.","ref":"CubDB.html#back_up/2"},{"type":"function","title":"CubDB.child_spec/1","doc":"Returns a specification to start this module under a supervisor. The default options listed in Supervisor are used.","ref":"CubDB.html#child_spec/1"},{"type":"function","title":"CubDB.clear/1","doc":"Deletes all entries, resulting in an empty database. The deletion is atomic, and is much more performant than deleating each entry manually. The operation respects all the guarantees of consistency of other concurrent operations. For example, if select2 was called before the call to clear/1 and is running concurrently, the select2 will still see all the entries. If a compaction is in progress when clear/1 is called, the compaction is halted, and a new one started immediately after. The new compaction should be very fast, as the database is empty as a result of the clear/1 call.","ref":"CubDB.html#clear/1"},{"type":"function","title":"CubDB.compact/1","doc":"Runs a database compaction. As write operations are performed on a database, its file grows. Occasionally, a compaction operation can be run to shrink the file to its optimal size. Compaction runs in the background and does not block operations. Only one compaction operation can run at any time, therefore if this function is called when a compaction is already running, it returns {:error, :pending_compaction} . When compacting, CubDB will create a new data file, and eventually switch to it and remove the old one as the compaction succeeds. For this reason, during a compaction, there should be enough disk space for a second copy of the database file. Compaction can create disk contention, so it should not be performed unnecessarily often.","ref":"CubDB.html#compact/1"},{"type":"function","title":"CubDB.compacting?/1","doc":"Returns true if a compaction operation is currently running, false otherwise.","ref":"CubDB.html#compacting?/1"},{"type":"function","title":"CubDB.current_db_file/1","doc":"Returns the path of the current database file. The current database file will change after a compaction operation. Example { :ok , db } = CubDB . start_link ( &quot;some/data/directory&quot; ) CubDB . current_db_file ( db ) #=&gt; &quot;some/data/directory/0.cub&quot;","ref":"CubDB.html#current_db_file/1"},{"type":"function","title":"CubDB.data_dir/1","doc":"Returns the path of the data directory, as given when the CubDB process was started. Example { :ok , db } = CubDB . start_link ( &quot;some/data/directory&quot; ) CubDB . data_dir ( db ) #=&gt; &quot;some/data/directory&quot;","ref":"CubDB.html#data_dir/1"},{"type":"function","title":"CubDB.delete/2","doc":"Deletes the entry associated to key from the database. If key was not present in the database, nothing is done.","ref":"CubDB.html#delete/2"},{"type":"function","title":"CubDB.delete_multi/2","doc":"Deletes multiple entries corresponding to the given keys all at once, atomically. The keys to be deleted are passed as a list.","ref":"CubDB.html#delete_multi/2"},{"type":"function","title":"CubDB.dirt_factor/1","doc":"Returns the dirt factor. The dirt factor is a number, ranging from 0 to 1, giving an indication about the amount of overhead disk space (or &quot;dirt&quot;) that can be cleaned up with a compaction operation. A value of 0 means that there is no overhead, so a compaction would have no benefit. The closer to 1 the dirt factor is, the more can be cleaned up in a compaction operation.","ref":"CubDB.html#dirt_factor/1"},{"type":"function","title":"CubDB.fetch/2","doc":"Fetches the value for the given key in the database, or returns :error if key is not present. If the database contains an entry with the given key and value value , it returns {:ok, value} . If key is not found, it returns :error .","ref":"CubDB.html#fetch/2"},{"type":"function","title":"CubDB.file_sync/1","doc":"Performs a fsync , forcing to flush all data that might be buffered by the OS to disk. Calling this function ensures that all writes up to this point are committed to disk, and will be available after a restart. If CubDB is started with the option auto_file_sync: true , calling this function is not necessary, as every write operation will be automatically flushed to the storage device. If this function is NOT called, the operative system will control when the file buffer is flushed to the storage device, which leads to better write performance, but might affect durability of recent writes in case of a sudden shutdown.","ref":"CubDB.html#file_sync/1"},{"type":"function","title":"CubDB.get/3","doc":"Gets the value associated to key from the database. If no value is associated with key , default is returned (which is nil , unless specified otherwise).","ref":"CubDB.html#get/3"},{"type":"function","title":"CubDB.get_and_update/3","doc":"Gets the value corresponding to key and updates it, in one atomic transaction. fun is called with the current value associated to key (or nil if not present), and must return a two element tuple: the result value to be returned, and the new value to be associated to key . fun may also return :pop , in which case the current value is deleted and returned. Note that in case the value to update returned by fun is the same as the original value, no write is performed to disk.","ref":"CubDB.html#get_and_update/3"},{"type":"function","title":"CubDB.get_and_update_multi/3","doc":"Gets and updates or deletes multiple entries in an atomic transaction. Gets all values associated with keys in keys_to_get , and passes them as a map of %{key =&gt; value} entries to fun . If a key is not found, it won't be added to the map passed to fun . Updates the database and returns a result according to the return value of fun . The function fun should return a tuple of three elements: {return_value, entries_to_put, keys_to_delete} , where return_value is an arbitrary value to be returned, entries_to_put is a map of %{key =&gt; value} entries to be written to the database, and keys_to_delete is a list of keys to be deleted. The read and write operations are executed as an atomic transaction, so they will either all succeed, or all fail. Note that get_and_update_multi/3 blocks other write operations until it completes. Example Assuming a database of names as keys, and integer monetary balances as values, and we want to transfer 10 units from &quot;Anna&quot; to &quot;Joy&quot; , returning their updated balance: { anna , joy } = CubDB . get_and_update_multi ( db , [ &quot;Anna&quot; , &quot;Joy&quot; ] , fn entries -&gt; anna = Map . get ( entries , &quot;Anna&quot; , 0 ) joy = Map . get ( entries , &quot;Joy&quot; , 0 ) if anna &lt; 10 , do : raise ( RuntimeError , message : &quot;Anna&#39;s balance is too low&quot; ) anna = anna - 10 joy = joy + 10 { { anna , joy } , %{ &quot;Anna&quot; =&gt; anna , &quot;Joy&quot; =&gt; joy } , [ ] } end ) Or, if we want to transfer all of the balance from &quot;Anna&quot; to &quot;Joy&quot; , deleting &quot;Anna&quot; 's entry, and returning &quot;Joy&quot; 's resulting balance: joy = CubDB . get_and_update_multi ( db , [ &quot;Anna&quot; , &quot;Joy&quot; ] , fn entries -&gt; anna = Map . get ( entries , &quot;Anna&quot; , 0 ) joy = Map . get ( entries , &quot;Joy&quot; , 0 ) joy = joy + anna { joy , %{ &quot;Joy&quot; =&gt; joy } , [ &quot;Anna&quot; ] } end )","ref":"CubDB.html#get_and_update_multi/3"},{"type":"function","title":"CubDB.get_multi/2","doc":"Gets multiple entries corresponding by the given keys all at once, atomically. The keys to get are passed as a list. The result is a map of key/value entries corresponding to the given keys. Keys that are not present in the database won't be in the result map. Example CubDB . put_multi ( db , a : 1 , b : 2 , c : nil ) CubDB . get_multi ( db , [ :a , :b , :c , :x ] ) # =&gt; %{a: 1, b: 2, c: nil}","ref":"CubDB.html#get_multi/2"},{"type":"function","title":"CubDB.halt_compaction/1","doc":"Stops a running compaction. If a compaction operation is running, it is halted, and the function returns :ok . Otherwise it returns {:error, :no_compaction_running} . If a new compaction is started (manually or automatically), it will start from scratch, the halted compaction is completely discarded. This function can be useful if one wants to make sure that no compaction operation is running in a certain moment, for example to perform some write-intensive workload without incurring in additional load. In this case one can pause auto compaction, and call halt_compaction/1 to stop any running compaction.","ref":"CubDB.html#halt_compaction/1"},{"type":"function","title":"CubDB.has_key?/2","doc":"Returns whether an entry with the given key exists in the database.","ref":"CubDB.html#has_key?/2"},{"type":"function","title":"CubDB.put/3","doc":"Writes an entry in the database, associating key to value . If key was already present, it is overwritten.","ref":"CubDB.html#put/3"},{"type":"function","title":"CubDB.put_and_delete_multi/3","doc":"Writes and deletes multiple entries all at once, atomically. Entries to put are passed as a map of %{key =&gt; value} or a list of {key, value} . Keys to delete are passed as a list of keys.","ref":"CubDB.html#put_and_delete_multi/3"},{"type":"function","title":"CubDB.put_multi/2","doc":"Writes multiple entries all at once, atomically. Entries are passed as a map of %{key =&gt; value} or a list of {key, value} .","ref":"CubDB.html#put_multi/2"},{"type":"function","title":"CubDB.put_new/3","doc":"Writes an entry in the database, associating key to value , only if key is not yet in the database. If key is already present, it does not change it, and returns {:error, :exists} .","ref":"CubDB.html#put_new/3"},{"type":"function","title":"CubDB.release_snapshot/1","doc":"Releases a snapshot when it is not needed anymore, releasing related resources This allows CubDB to perform cleanup operations after compaction that are otherwise blocked by the snapshot. When creating a snapshot with a timeout, it is not necessary to call release_snapshot/1 , as it will be automatically released after the timeout elapses. When getting a snapshot with a timeout of :infinity though, one has to manually call release_snapshot/1 once the snapshot is not needed anymore. In most cases, using with_snapshot/2 is a better alternative to manually calling snapshot/2 and release_snapshot/1","ref":"CubDB.html#release_snapshot/1"},{"type":"function","title":"CubDB.select/2","doc":"Selects a range of entries from the database, returning a lazy stream. The returned lazy stream can be filtered, mapped, and transformed with standard functions in the Stream and Enum modules. The actual database read is deferred to when the stream is iterated or evaluated. Options The min_key and max_key specify the range of entries that are selected. By default, the range is inclusive, so all entries that have a key greater or equal than min_key and less or equal then max_key are selected: # Select all entries where &quot;a&quot; &lt;= key &lt;= &quot;d&quot; CubDB . select ( db , min_key : &quot;b&quot; , max_key : &quot;d&quot; ) The range boundaries can be excluded by setting min_key_inclusive or max_key_inclusive to false : # Select all entries where &quot;a&quot; &lt;= key &lt; &quot;d&quot; CubDB . select ( db , min_key : &quot;b&quot; , max_key : &quot;d&quot; , max_key_inclusive : false ) Any of :min_key and :max_key can be omitted, to leave the range open-ended. # Select entries where key &lt;= &quot;a&quot; CubDB . select ( db , max_key : &quot;a&quot; ) Since nil is a valid key, setting min_key or max_key to nil does NOT leave the range open ended: # Select entries where nil &lt;= key &lt;= &quot;a&quot; CubDB . select ( db , min_key : nil , max_key : &quot;a&quot; ) The reverse option, when set to true, causes the entries to be selected and traversed in reverse order. This is more efficient than selecting them in normal ascending order and then reversing the resulting collection. Note that, when selecting a key range, specifying min_key and/or max_key is more performant than using functions in Enum or Stream to filter out entries out of range, because min_key and max_key avoid loading unnecessary entries from disk entirely. Examples To select all entries with keys between :a and :c as a stream of {key, value} entries we can do: entries = CubDB . select ( db , min_key : :a , max_key : :c ) Since select/2 returns a lazy stream, at this point nothing has been fetched from the database yet. We can turn the stream into a list, performing the actual query: Enum . to_list ( entries ) If we want to get all entries with keys between :a and :c , with :c excluded, we can do: entries = CubDB . select ( db , min_key : :a , max_key : :c , max_key_inclusive : false ) |&gt; Enum . to_list ( ) To select the last 3 entries, we can do: entries = CubDB . select ( db , reverse : true ) |&gt; Enum . take ( 3 ) If we want to obtain the sum of the first 10 positive numeric values associated to keys from :a to :f , we can do: sum = CubDB . select ( db , min_key : :a , max_key : :f ) |&gt; Stream . map ( fn { _key , value } -&gt; value end ) # map values |&gt; Stream . filter ( fn n -&gt; is_number ( n ) and n &gt; 0 end ) # only positive numbers |&gt; Stream . take ( 10 ) # take only the first 10 entries in the range |&gt; Enum . sum ( ) # sum the selected values Using functions from the Stream module for mapping and filtering ensures that we do not fetch unnecessary items from the database. In the example above, for example, after fetching the first 10 entries satisfying the filter condition, no further entry is fetched from the database.","ref":"CubDB.html#select/2"},{"type":"function","title":"CubDB.set_auto_compact/2","doc":"Configures whether to perform automatic compaction, and how. If set to false , no automatic compaction is performed. If set to true , auto-compaction is performed, following a write operation, if at least 100 write operations occurred since the last compaction, and the dirt factor is at least 0.25. These values can be customized by setting the auto_compact option to {min_writes, min_dirt_factor} . It returns :ok , or {:error, reason} if setting is invalid. Compaction is performed in the background and does not block other operations, but can create disk contention, so it should not be performed unnecessarily often. When writing a lot into the database, such as when importing data from an external source, it is advisable to turn off auto compaction, and manually run compaction at the end of the import.","ref":"CubDB.html#set_auto_compact/2"},{"type":"function","title":"CubDB.set_auto_file_sync/2","doc":"Configures whether to automatically force file sync upon each write operation. If set to false , no automatic file sync is performed. That improves write performance, but leaves to the operative system the decision of when to flush disk buffers. This means that there is the possibility that recent writes might not be durable in case of a sudden machine shutdown. In any case, atomicity of multi operations is preserved, and partial writes will not corrupt the database. If set to true , the file buffer will be forced to flush upon every write operation, ensuring durability even in case of sudden machine shutdowns, but decreasing write performance.","ref":"CubDB.html#set_auto_file_sync/2"},{"type":"function","title":"CubDB.size/1","doc":"Returns the number of entries present in the database.","ref":"CubDB.html#size/1"},{"type":"function","title":"CubDB.snapshot/2","doc":"Returns a snapshot of the database in its current state. Note It is usually better to use with_snapshot/2 instead of snapshot/2 , as the former automatically manages the snapshot life cycle, even in case of crashes. A snapshot is an immutable, read-only representation of the database at a specific point in time. Getting a snapshot is basically zero-cost: nothing needs to be copied or written, apart from some small in-memory bookkeeping. The only cost of a snapshot is that it delays cleanup of old files after compaction for as long as it is in use. For this reason, a snapshot has a timeout, configurable as the optional second argument of snapshot/2 (defaulting to 5000, or 5 seconds, if not specified). After such timeout elapses, the snapshot cannot be used anymore, and any pending cleanup is performed. It is possible to pass :infinity as the timeout, but then one must manually call release_snapshot/1 to release the snapshot after use. Using with_snapshot/1 is often a better alternative to snapshot/2 , as it does not require to choose an arbitrary timeout, and automatically ensures that the the snapshot is released after use, even in case of a crash. After obtaining a snapshot, it is possible to read from it using the functions in CubDB.Snapshot , which work the same way as the functions in CubDB with the same name, such as CubDB.Snapshot.get/3 , CubDB.Snapshot.get_multi/2 , CubDB.Snapshot.fetch/2 , CubDB.Snapshot.has_key?/2 , CubDB.Snapshot.select/2 . It is not possible to write on a snapshot. Example CubDB . put ( db , :a , 123 ) snap = CubDB . snapshot ( db ) CubDB . put ( db , :a , 0 ) # Getting a value from the snapshot returns the value of the entry at the # time the snapshot was obtained, even if the entry has changed in the # meanwhile CubDB.Snapshot . get ( snap , :a ) # =&gt; 123 # Getting the same value from the database returns the latest value CubDB . get ( db , :a ) # =&gt; 0","ref":"CubDB.html#snapshot/2"},{"type":"function","title":"CubDB.start/1","doc":"Starts the CubDB database without a link. See start_link/2 for more information about options.","ref":"CubDB.html#start/1"},{"type":"function","title":"CubDB.start/2","doc":"","ref":"CubDB.html#start/2"},{"type":"function","title":"CubDB.start_link/1","doc":"Starts the CubDB database process linked to the current process. The argument is a keyword list of options: data_dir : the directory path where the database files will be stored. This option is required. If the directory does not exist, it will be created. Only one CubDB instance can run per directory, so if you run several databases, they should each use their own separate data directory. auto_compact : whether to perform compaction automatically. It defaults to true . See set_auto_compact/2 for the possible values auto_file_sync : whether to force flush the disk buffer on each write. It defaults to true . If set to false , write performance is faster, but durability of writes is not strictly guaranteed. See set_auto_file_sync/2 for details. GenServer options like name and timeout can also be given, and are forwarded to GenServer.start_link/3 as the third argument. If only the data_dir is specified, it is possible to pass it as a single string argument. Examples # Passing only the data dir { :ok , db } = CubDB . start_link ( &quot;some/data/dir&quot; ) # Passing data dir and other options { :ok , db } = CubDB . start_link ( data_dir : &quot;some/data/dir&quot; , auto_compact : true , name : :db )","ref":"CubDB.html#start_link/1"},{"type":"function","title":"CubDB.start_link/2","doc":"","ref":"CubDB.html#start_link/2"},{"type":"function","title":"CubDB.stop/3","doc":"Synchronously stops the CubDB database. See GenServer.stop/3 for details.","ref":"CubDB.html#stop/3"},{"type":"function","title":"CubDB.transaction/2","doc":"Starts a write transaction, passes it to the given function, and commits or cancels it depending on the return value. The transaction blocks other writers until the function returns, but does not block concurrent readers. When the need is to only read inside a transaction, and not perform any write, using a snapshot is a better choice, as it does not block writers (see with_snapshot/2 ). The module CubDB.Tx contains functions to perform read and write operations within the transaction. The function fun is called with the transaction as argument, and should return {:commit, tx, results} to commit the transaction tx and return result , or {:cancel, result} to cancel the transaction and return result . If an exception is raised, or a value thrown, or the process exits while inside of a transaction, the transaction is cancelled. Only use CubDB.Tx functions to write when inside a transaction (like CubDB.Tx.put/3 or CubDB.Tx.delete/2 ). Using functions in the CubDB module to perform a write when inside a transaction (like CubDB.put/3 or CubDB.delete/2 ) raises an exception. Note that write functions in CubDB.Tx have a functional API: they return a modified transaction rather than mutating it in place. The transaction value passed to fun should not be used outside of the function. Example: Suppose the keys :a and :b map to balances, and we want to transfer 5 from :a to :b , if :a has enough balance: CubDB . transaction ( db , fn tx -&gt; a = CubDB.Tx . get ( tx , :a ) b = CubDB.Tx . get ( tx , :b ) if a &gt;= 5 do tx = CubDB.Tx . put ( tx , :a , a - 5 ) tx = CubDB.Tx . put ( tx , :b , b + 5 ) { :commit , tx , :ok } else { :cancel , :not_enough_balance } end end ) The read functions in CubDB.Tx read the in-transaction state, as opposed to the live database state, so they see writes performed inside the transaction even before they are committed: # Assuming we start from an empty database CubDB . transaction ( db , fn tx -&gt; tx = CubDB.Tx . put ( tx , :a , 123 ) # CubDB.Tx.get sees the in-transaction value CubDB.Tx . get ( tx , :a ) # =&gt; 123 # CubDB.get instead does not see the uncommitted write CubDB . get ( db , :a ) # =&gt; nil { :commit , tx , nil } end ) # After the transaction is committed, CubDB.get sees the write CubDB . get ( db , :a ) # =&gt; 123","ref":"CubDB.html#transaction/2"},{"type":"function","title":"CubDB.update/4","doc":"Updates the entry corresponding to key using the given function. If key is present in the database, fun is invoked with the corresponding value , and the result is set as the new value of key . If key is not found, initial is inserted as the value of key .","ref":"CubDB.html#update/4"},{"type":"function","title":"CubDB.with_snapshot/2","doc":"Calls fun passing a snapshot, and automatically releases the snapshot when the function returns It returns the value returned by the function fun . A snapshot is an immutable, read-only representation of the database at a specific point in time, isolated from writes. It is basically zero-cost: nothing needs to be copied or written, apart from some small in-memory bookkeeping. Calling with_snapshot/2 is equivalent to obtaining a snapshot with snapshot/2 using a timeout of :infinity , calling fun , then manually releasing the snapshot with release_snapshot/1 , but with_snapshot/2 automatically manages the snapshot life cycle, also in case an exception is raised, a value is thrown, or the process exists. This makes with_snapshot/2 usually a better choice than snapshot/2 . After obtaining a snapshot, it is possible to read from it using the functions in CubDB.Snapshot , which work the same way as the functions in CubDB with the same name, such as CubDB.Snapshot.get/3 , CubDB.Snapshot.get_multi/2 , CubDB.Snapshot.fetch/2 , CubDB.Snapshot.has_key?/2 , CubDB.Snapshot.select/2 . It is not possible to write on a snapshot. Example Assume that we have two entries in the database, and the key of the second entry depends on the value of the first (so the value of the first entry &quot;points&quot; to the other entry). In this case, we want to get both entries from the same snapshot, to avoid inconsistencies due to concurrent writes. Here's how that can be done with with_snapshot/2 : { x , y } = CubDB . with_snapshot ( db , fn snap -&gt; x = CubDB.Snapshot . get ( snap , :x ) y = CubDB.Snapshot . get ( snap , x ) { x , y } end )","ref":"CubDB.html#with_snapshot/2"},{"type":"type","title":"CubDB.auto_compact/0","doc":"","ref":"CubDB.html#t:auto_compact/0"},{"type":"type","title":"CubDB.data_dir_option/0","doc":"","ref":"CubDB.html#t:data_dir_option/0"},{"type":"type","title":"CubDB.entry/0","doc":"","ref":"CubDB.html#t:entry/0"},{"type":"type","title":"CubDB.key/0","doc":"","ref":"CubDB.html#t:key/0"},{"type":"type","title":"CubDB.option/0","doc":"","ref":"CubDB.html#t:option/0"},{"type":"type","title":"CubDB.select_option/0","doc":"","ref":"CubDB.html#t:select_option/0"},{"type":"type","title":"CubDB.server/0","doc":"","ref":"CubDB.html#t:server/0"},{"type":"type","title":"CubDB.value/0","doc":"","ref":"CubDB.html#t:value/0"},{"type":"module","title":"CubDB.Snapshot","doc":"The CubDB.Snapshot module contains functions to read from snapshots obtained with CubDB.with_snapshot/2 or CubDB.snapshot/2 . The functions in this module mirror the ones with the same name in the CubDB module, but work on snapshots instead of on the live database.","ref":"CubDB.Snapshot.html"},{"type":"function","title":"CubDB.Snapshot.back_up/2","doc":"Creates a backup of the snapshot into the target directory path It works the same as CubDB.back_up/2 , but works on a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#back_up/2"},{"type":"function","title":"CubDB.Snapshot.fetch/2","doc":"Fetches the value for the given key from the snapshot, or returns :error if key is not present. If the snapshot contains an entry with the given key and value value , it returns {:ok, value} . If key is not found, it returns :error . It works the same as CubDB.fetch/2 , but reads from a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#fetch/2"},{"type":"function","title":"CubDB.Snapshot.get/3","doc":"Gets the value associated to key from the snapshot. If no value is associated with key , default is returned (which is nil , unless specified otherwise). It works the same as CubDB.get/3 , but reads from a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#get/3"},{"type":"function","title":"CubDB.Snapshot.get_multi/2","doc":"Gets multiple entries corresponding by the given keys from the snapshot all at once, atomically. It works the same as CubDB.get_multi/2 , but reads from a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#get_multi/2"},{"type":"function","title":"CubDB.Snapshot.has_key?/2","doc":"Returns whether an entry with the given key exists in the snapshot. It works the same as CubDB.has_key?/2 , but reads from a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#has_key?/2"},{"type":"function","title":"CubDB.Snapshot.select/2","doc":"Selects a range of entries from the snapshot, returning a lazy stream. The lazy stream should be evaluated while the snapshot is still valid, or a RuntimeError will be raised. It works the same and accepts the same options as CubDB.select/2 , but reads from a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#select/2"},{"type":"function","title":"CubDB.Snapshot.size/1","doc":"Returns the number of entries present in the snapshot. It works the same as CubDB.size/1 , but works on a snapshot instead of the live database.","ref":"CubDB.Snapshot.html#size/1"},{"type":"type","title":"CubDB.Snapshot.t/0","doc":"","ref":"CubDB.Snapshot.html#t:t/0"},{"type":"module","title":"CubDB.Tx","doc":"The CubDB.Tx module contains functions to read or write within atomic transactions created with CubDB.transaction/2 . The functions in this module mirror the ones with the same name in the CubDB module, but work on transactions instead of on the live database.","ref":"CubDB.Tx.html"},{"type":"function","title":"CubDB.Tx.clear/1","doc":"Deletes all entries, resulting in an empty database. The deletion is much more performant than deleating each entry manually. It works like CubDB.clear/1 , but on a transaction. If a compaction is in progress when clear/1 is called, and the transaction is committed, the compaction is halted, and a new one started immediately after. The new compaction should be very fast, as the database is empty as a result of the clear/1 call. Like other functions in this module, it does not mutate the transaction in place, and instead returns a modified transaction.","ref":"CubDB.Tx.html#clear/1"},{"type":"function","title":"CubDB.Tx.delete/2","doc":"Deletes the entry associated to key from the transaction. If key was not present, nothing is done. Like other functions in this module, it does not mutate the transaction in place, and instead returns a modified transaction.","ref":"CubDB.Tx.html#delete/2"},{"type":"function","title":"CubDB.Tx.fetch/2","doc":"Fetches the value for the given key from the transaction, or returns :error if key is not present. If the snapshot contains an entry with the given key and value value , it returns {:ok, value} . If key is not found, it returns :error . It works the same as CubDB.fetch/2 , but reads from a transaction instead of the live database.","ref":"CubDB.Tx.html#fetch/2"},{"type":"function","title":"CubDB.Tx.get/3","doc":"Gets the value associated to key from the transaction. If no value is associated with key , default is returned (which is nil , unless specified otherwise). It works the same as CubDB.get/3 , but reads from a transaction instead of the live database.","ref":"CubDB.Tx.html#get/3"},{"type":"function","title":"CubDB.Tx.has_key?/2","doc":"Returns whether an entry with the given key exists in the transaction. It works the same as CubDB.has_key?/2 , but reads from a transaction instead of the live database.","ref":"CubDB.Tx.html#has_key?/2"},{"type":"function","title":"CubDB.Tx.put/3","doc":"Writes an entry in the transaction, associating key to value . If key was already present, it is overwritten. Like other functions in this module, it does not mutate the transaction in place, and instead returns a modified transaction.","ref":"CubDB.Tx.html#put/3"},{"type":"function","title":"CubDB.Tx.put_new/3","doc":"Writes an entry in the transaction, associating key to value , only if key is not yet present. If key is already present, it does not change it, and returns {:error, :exists} . Like other functions in this module, it does not mutate the transaction in place, and instead returns a modified transaction.","ref":"CubDB.Tx.html#put_new/3"},{"type":"function","title":"CubDB.Tx.refetch/3","doc":"If key was not written since the point in time represented by snapshot , returns :unchanged . Otherwise, behaves like fetch/2 . Checking if the key was written is done without fetching the whole entry, but instead only checking index pages. This makes refetch/3 useful as a performance optimization in cases when one needs to verify if an entry changed with respect to a snapshot: using refetch/3 can save some disk access if CubDB is able to determine that the key was not written since the snapshot. In some situations, such as when the entry is not present in the database, or when a compaction completed after snapshot , the function cannot determine if the entry was written or not since snapshot , and therefore fetches it. In other words, refetch/3 is a performance optimization to save some disk reads when possible, but it might fetch an entry even if it technically did not change. Example The function refetch/3 can be useful when implementing optimistic concurrency control. Suppose, for example, that computing the updated value of some entry is a slow operation. In order to avoid holding a transaction open for too long, one could compute the update outside of the transaction, and then check if the value on which the update was computed is still the same: if so, commit the update, otherwise perform the computation again. This kind of optimistic concurrency control can use refetch/3 to avoid reading the value from disk when nothing has changed: def update_optimistically ( db , key ) do outcome = CubDB . with_snapshot ( db , fn snap -&gt; { :ok , value } = CubDB.Snapshot . fetch ( snap , key ) # Perform the slow calculation outside of the transaction new_value = some_slow_calculation ( value ) # In a transaction, check if the value changed, and update it if not write_if_unchanged ( db , key , value , snap ) end ) # Depending on the outcome, return or recompute case outcome do :recompute -&gt; update_optimistically ( db , key ) :ok -&gt; :ok end end defp write_if_unchanged ( db , key , value , snap ) do CubDB . transaction ( db , fn tx -&gt; # Check if the value changed in the meanwhile case CubDB.Tx . refetch ( tx , key , snap ) do :unchanged -&gt; # The entry was not written since we last read it. Commit the # new value and return :ok { :commit , CubDB.Tx . put ( tx , key , new_value ) , :ok } { :ok , ^ value } -&gt; # The entry was written, but its value did not change. Commit the # new value and return :ok { :commit , CubDB.Tx . put ( tx , key , new_value ) , :ok } _ -&gt; # The entry changed since we last read it, cancel the # transaction and return :recompute { :cancel , :recompute } end end ) end","ref":"CubDB.Tx.html#refetch/3"},{"type":"function","title":"CubDB.Tx.select/2","doc":"Selects a range of entries from the transaction, returning a lazy stream. The lazy stream should be evaluated within the transaction scope, or a RuntimeError will be raised. It works the same and accepts the same options as CubDB.select/2 , but reads from a transaction instead of the live database.","ref":"CubDB.Tx.html#select/2"},{"type":"function","title":"CubDB.Tx.size/1","doc":"Returns the number of entries present in the transaction. It works the same as CubDB.size/1 , but works on a transaction instead of the live database.","ref":"CubDB.Tx.html#size/1"},{"type":"type","title":"CubDB.Tx.t/0","doc":"","ref":"CubDB.Tx.html#t:t/0"},{"type":"extras","title":"Changelog","doc":"For versions before v1.0.0 , the changes are reported in the GitHub releases . Since v1.0.0 , CubDB follows semantic versioning , and reports changes here.","ref":"changelog.html"},{"type":"extras","title":"Changelog - v2.0.2 (2023-01-01)","doc":"Bug fixes: Better exception in case of file errors (such as :efbig )","ref":"changelog.html#v2-0-2-2023-01-01"},{"type":"extras","title":"Changelog - v2.0.1 (2022-08-30)","doc":"Various bug fixes: Avoid unnecessary commits in the compaction catching up phase More informative error when running out of disk space Avoid leaking processes when stopping CubDB during a compaction","ref":"changelog.html#v2-0-1-2022-08-30"},{"type":"extras","title":"Changelog - v2.0.0 (2022-06-28)","doc":"Version 2.0.0 brings better concurrency, atomic transactions with arbitrary operations, zero cost read-only snapshots, database backup, and more, all with a simpler and more scalable internal architecture. Refer to the upgrade guide for how to upgrade from previous versions. [breaking] The functions CubDB.get_and_update/3 , CubDB.get_and_update_multi/3 , and CubDB.select/2 now return directly result , instead of a {:ok, result} tuple. [breaking] CubDB.get_and_update_multi/4 does not take an option argument anymore, making it CubDB.get_and_update_multi/3 . The only available option used to be :timeout , which is not supported anymore. [breaking] Remove the :timeout option on CubDB.select/2 . This is part of a refactoring and improvement that moves read operations from an internally spawned Task to the client process. This makes the :timeout option unnecessary: by stopping the process calling CubDB , any running read operation by that process is stopped. [breaking] CubDB.select/2 now returns a lazy stream that can be used with functions in Enum and Stream . This makes the :pipe and :reduce options unnecessary, so those options were removed. Add CubDB.snapshot/2 , CubDB.with_snapshot/2 and CubDB.release_snapshot/1 to get zero cost read-only snapshots of the database. The functions in CubDB.Snapshot allow to read from a snapshot. Add CubDB.transaction/2 to perform multiple write (and read) operations in a single atomic transaction. The functions in CubDB.Tx allow to read and write inside a transaction. Add CubDB.back_up/2 to produce a database backup. The backup process does not block readers or writers, and is isolated from concurrent writes. Add CubDB.halt_compaction/1 to stop any running compaction operation Add CubDB.compacting?/1 to check if a compaction is currently running Move read and write operations to the caller process as opposed to the CubDB server process. Improve concurrency of read operations while writing","ref":"changelog.html#v2-0-0-2022-06-28"},{"type":"extras","title":"Changelog - v1.1.0 (2021-10-14)","doc":"Add clear/1 function to atomically delete all entries in the database","ref":"changelog.html#v1-1-0-2021-10-14"},{"type":"extras","title":"Changelog - v1.0.0 (2021-06-24)","doc":"Breaking changes from v0.17.0: Better defaults: auto_file_sync now defaults to true (slower but durable) auto_compact now defaults to true Functions select/2 and get_and_update_multi/4 now take the timeout as an option instead of an additional argument Other changes from v0.17.0: Better internal handling of timeouts that ensures cleanup of resources on the callee side Added put_and_delete_multi/3 to atomically put and delete entries Added put_new/3 to put an entry only if the key does not exist yet More efficient implementation of put_multi/2 and delete_multi/2 Function get_multi/2 does not block writers Fix race condition during compaction Function get_and_update/3 avoids unnecessary disk writes (including the transaction header) when the value is unchanged Remove caller timeout on put_and_delete_multi/3 and put_multi/2 , consistently with the other functions. Remove default GenServer timeouts Fix process (and file descriptor) leak upon compaction Fix cubdb_file?/1 regexp, making it stricter","ref":"changelog.html#v1-0-0-2021-06-24"},{"type":"extras","title":"License","doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS Copyright 2023 Luca Ongaro Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html"},{"type":"extras","title":"Overview","doc":"CubDB is an embedded key-value database for the Elixir language. It is designed for robustness, and for minimal need of resources. Head to the API reference for usage details, or read the Frequently Asked Questions and the How To section for more information.","ref":"readme.html"},{"type":"extras","title":"Overview - Features","doc":"Both keys and values can be any Elixir (or Erlang) term. Basic get , put , and delete operations, selection of ranges of entries sorted by key with select . Atomic, Consistent, Isolated, Durable (ACID) transactions. Multi version concurrency control (MVCC) allowing concurrent read operations, that do not block nor are blocked by writes. Unexpected shutdowns or crashes won't corrupt the database or break atomicity of transactions. Manual or automatic compaction to reclaim disk space. To ensure consistency, performance, and robustness to data corruption, CubDB database file uses an append-only, immutable B-tree data structure. Entries are never changed in-place, and read operations are performed on zero cost immutable snapshots.","ref":"readme.html#features"},{"type":"extras","title":"Overview - Usage","doc":"Start CubDB by specifying a directory for its database file (if not existing, it will be created): { :ok , db } = CubDB . start_link ( data_dir : &quot;my/data/directory&quot; ) Important Avoid starting multiple CubDB processes on the same data directory. Only one CubDB process should use a specific data directory at any time. get , put , and delete operations work as you probably expect: CubDB . put ( db , :foo , &quot;some value&quot; ) #=&gt; :ok CubDB . get ( db , :foo ) #=&gt; &quot;some value&quot; CubDB . delete ( db , :foo ) #=&gt; :ok CubDB . get ( db , :foo ) #=&gt; nil Multiple operations can be performed atomically with the transaction function and the CubDB.Tx module: # Swapping `:a` and `:b` atomically: CubDB . transaction ( db , fn tx -&gt; a = CubDB.Tx . get ( tx , :a ) b = CubDB.Tx . get ( tx , :b ) tx = CubDB.Tx . put ( tx , :a , b ) tx = CubDB.Tx . put ( tx , :b , a ) { :commit , tx , :ok } end ) #=&gt; :ok Alternatively, it is possible to use put_multi , delete_multi , and the other [...]_multi functions, which also guarantee atomicity: CubDB . put_multi ( db , [ a : 1 , b : 2 , c : 3 , d : 4 , e : 5 , f : 6 , g : 7 , h : 8 ] ) #=&gt; :ok Range of entries sorted by key are retrieved using select : CubDB . select ( db , min_key : :b , max_key : :e ) |&gt; Enum . to_list ( ) #=&gt; [b: 2, c: 3, d: 4, e: 5] The select function can select entries in normal or reverse order, and returns a lazy stream, so one can use functions in the Stream and Enum modules to map, filter, and transform the result, only fetching from the database the relevant entries: # Take the sum of the last 3 even values: CubDB . select ( db , reverse : true ) # select entries in reverse order |&gt; Stream . map ( fn { _key , value } -&gt; value end ) # discard the key and keep only the value |&gt; Stream . filter ( fn value -&gt; is_integer ( value ) &amp;&amp; Integer . is_even ( value ) end ) # filter only even integers |&gt; Stream . take ( 3 ) # take the first 3 values |&gt; Enum . sum ( ) # sum the values #=&gt; 18 Read-only snapshots are useful when one needs to perform several reads or selects, ensuring isolation from concurrent writes, but without blocking them. When nothing needs to be written, using a snapshot is preferable to using a transaction, because it will not block writes. Snapshots come at no cost: nothing is actually copied or written on disk or in memory, apart from some small internal bookkeeping. After obtaining a snapshot with with_snapshot , one can read from it using the functions in the CubDB.Snapshot module: # the key of y depends on the value of x, so we ensure consistency by getting # both entries from the same snapshot, isolating from the effects of concurrent # writes { x , y } = CubDB . with_snapshot ( db , fn snap -&gt; x = CubDB.Snapshot . get ( snap , :x ) y = CubDB.Snapshot . get ( snap , x ) { x , y } end ) The functions that read multiple entries like get_multi , select , etc. are internally using a snapshot, so they always ensure consistency and isolation from concurrent writes, implementing multi version concurrency control (MVCC). For more details, read the API documentation .","ref":"readme.html#usage"},{"type":"extras","title":"Overview - Installation","doc":"CubDB can be installed by adding :cubdb to your list of dependencies in mix.exs : def deps do [ { :cubdb , &quot;~&gt; 2.0.2&quot; } ] end","ref":"readme.html#installation"},{"type":"extras","title":"Overview - Acknowledgement","doc":"The file data structure used by CubDB is inspired by CouchDB . A big thanks goes to the CouchDB maintainers for the readable codebase and extensive documentation.","ref":"readme.html#acknowledgement"},{"type":"extras","title":"Overview - Copyright and License","doc":"Copyright 2022 Luca Ongaro Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"readme.html#copyright-and-license"},{"type":"extras","title":"Frequently Asked Questions","doc":"","ref":"faq.html"},{"type":"extras","title":"Frequently Asked Questions - Why should I use CubDB?","doc":"If your Elixir application needs an embedded database (which is a database that run inside your application, as opposed to one that runs as a separate software), then CubDB might be a great fit: its API is simple and idiomatic, you can supervise it as any other Elixir process, and it runs wherever Elixir runs, with no need to cross-compile native code. You can think about CubDB as an Elixir collection, like a Map or List , but one that is stored to disk and persistent to restarts. The typical use-case for CubDB is data storage on single-instance applications or embedded software (for example, CubDB is a great fit for Nerves projects). In those contexts, CubDB is typically used for things like: Persisting configuration and preferences Data logging and storage of metrics or time series Local database for application data","ref":"faq.html#why-should-i-use-cubdb"},{"type":"extras","title":"Frequently Asked Questions - How does it compare with X?","doc":"CubDB is not jealous: it does its job well, without claiming to be better than others. There are many other great alternatives for storing data, and depending on your use-case they might be a better fit. That said, here are some reasons why you might choose CubDB over other popular alternatives: ETS (Erlang Term Store), like CubDB , can store arbitrary Elixir/Erlang terms, and comes by default with the Erlang runtime. Differently from CubDB though, ETS stores values in memory, so they are lost in case of a restart. DETS is similar to ETS, but it persists values on disk. Compared to CubDB though, it does not support sorted collections. Also, you might find its API more convoluted that CubDB . Mnesia is a distributed database system that comes with the Erlang runtime. It can be distributed, and enforces a schema. If on one hand it can make a great heavy-duty distributed database system, it is substantially more complicated to use than CubDB for embedded use cases. SQLite , LevelDB , LMDB , etc. are all great general-purpose embedded databases. They support a variety of features, and have wider adoption than CubDB . Because they are not &quot;Elixir native&quot; though, you need a library to interact with them from Elixir, and they generally feel less &quot;idiomatic&quot; than CubDB when used from an Elixir application. In some cases (like when writing software for embedded devices), it can be more complicated to install and cross-compile them for your architecture, and some libraries might not play well with your supervision strategies (NIF libraries are fast, but will crash the Erlang VM if the native executable crash). Plain files can be used to store simple data. Compared to CubDB though, they don't offer efficient key/value access or sorted collections, and they are subject to data corruption if a sudden power loss happens mid-write. CubDB is designed to be efficient and robust, so a power loss won't corrupt its data, or break atomicity. If your use case is such that one of those alternatives is a better fit, by all means go for it: CubDB won't be sad :) But if CubDB fits your use case, you will enjoy its native Elixir feel and simple but versatile model.","ref":"faq.html#how-does-it-compare-with-x"},{"type":"extras","title":"Frequently Asked Questions - What is compaction?","doc":"CubDB uses an append-only B-tree data structure: each change to CubDB is appended to the data file, instead of modifying the existing data in-place. This is efficient and robust: writing at the end of a file is faster than &quot;jumping around&quot;, and should something go wrong in the middle of a write (say, a power failure), no data is destroyed by a partial overwrite, so CubDB is able to recover upon restart. The drawback of this approach though, is that the data file will keep growing as you write to the database, even when you update or delete existing values. Performance of read and write operations is not affected by the file size, but space utilization can be optimized: old entries that are not &quot;reachable&quot; are still in the data file, making it larger than it needs to be. Compaction is the operation through which CubDB &quot;cleans up&quot; and makes its data file compact and space-efficient again. Other databases have similar processes (for example, SQLite calls it &quot;vacuuming&quot;). During a compaction operation, CubDB creates a new file, and transfers to it the current entries, without the stale data. When all the data is transferred, including entries written after the compaction started, CubDB switches to use the new compacted file and removes the old, obsolete one. The compaction process is performed in the background, without blocking read or write operations on the active data file. Should a compaction operation be interrupted by a shutdown or an application crash, no data is lost: the old data file is still up to date and active, and upon the next compaction the file left over by the interrupted compaction is removed and a new one created.","ref":"faq.html#what-is-compaction"},{"type":"extras","title":"Frequently Asked Questions - Should I use auto compaction?","doc":"Usually, letting CubDB perform compaction automatically is the most convenient choice: CubDB will keep track of its &quot;dirt factor&quot;, which is an indicator of how much overhead could be shaved off by a compaction, and clean up when necessary. In some situations though, it can be advisable to avoid auto compaction and compact your database manually. One example is if you are performing a one-off data import: in this case, it makes sense to import all data, and manually trigger a compaction only afterwards. This reduces disk contention during the import. Note that you can turn auto compaction on or off at runtime with CubDB.set_auto_compact/2 .","ref":"faq.html#should-i-use-auto-compaction"},{"type":"extras","title":"Frequently Asked Questions - What does file sync mean?","doc":"When you write to a file, your operative system usually buffers writes in memory, and actually writes them in the file system only later. This makes write operations faster: file system access is expensive, and buffering batches together several writes in one single operation. The drawback is that, should a power failure happen before the buffer is written to the file system, data that was held in the buffer might be lost. When you want to make sure that data is safe in the file system, you have to tell the operative system to &quot;file sync&quot;, which means to flush the buffer to disk. With CubDB , you can chose to automatically sync each write operation, or to manually sync when you prefer. If you need faster write performance, but you are ok with a small chance of losing recently written data in case of a power loss (for example if you use CubDB to log large quantity of data), you might choose to not perform automatic file sync. Note that the operative system will still sync data periodically, and upon clean shutdown. If instead you want to be sure that data that was successfully written by CubDB won't be lost, even in case of a power failure, you should use the auto file sync option: write operations will be sensibly slower, but each write will be safe on disk by the time the write operation completes. Even with auto file sync turned off, power failures won't corrupt the database or break atomicity. Whether to file sync or not is therefore a trade off between durability and write performance, and does not affect other semantics.","ref":"faq.html#what-does-file-sync-mean"},{"type":"extras","title":"How To","doc":"","ref":"howto.html"},{"type":"extras","title":"How To - Store multiple collections","doc":"The ability of CubDB to retrieve ranges of data sorted by key makes it ideal to store sorted collections, sort of like tables in a relational database. Different collections can be stored in the same database, or in separate ones. The main reason to use separate CubDB databases is if you need different configuration options for each database (like auto compaction or file sync). Otherwise, it is usually simpler and more efficient to use a single database: read operations in CubDB execute concurrently, and while write operations on the same database are serialized, using multiple databases means losing the advantage of the append-only nature of CubDB , resulting in often slower patterns of file access. One common way to store separate collections in the same database is to use composite keys, usually tuples, leveraging the fact that the Erlang runtime defines a total ordering of all terms . Say that you want to store two collections: people and articles. You can structure your keys to be tuples like {:people, person_id} for people, and {:articles, article_id} for articles: # Add a few people: :ok = CubDB . put ( db , { :people , 1 } , %{ first_name : &quot;Margaret&quot; , last_name : &quot;Hamilton&quot; } ) :ok = CubDB . put ( db , { :people , 2 } , %{ first_name : &quot;Alan&quot; , last_name : &quot;Turing&quot; } ) # Add a few articles: :ok = CubDB . put ( db , { :articles , 1 } , %{ title : &quot;Spaceship Guidance made easy&quot; , text : &quot;...&quot; } ) :ok = CubDB . put ( db , { :articles , 2 } , %{ title : &quot;Morphogenesis for the uninitiated&quot; , text : &quot;...&quot; } ) We used positive integers as IDs in our example, but you can really use anything you want. Getting a specific person or article by ID is trivial: person = CubDB . get ( db , { :people , 1 } ) article = CubDB . get ( db , { :articles , 2 } ) Selecting all members of a collection without selecting also other collections can be easily done by leveraging the fact that tuples are compared element by element. Therefore, here is how you can select all members of a specific collection: # Select all people { :ok , people_wth_keys } = CubDB . select ( db , min_key : { :people , 0 } , max_key : { :people , nil } ) # Select all articles { :ok , articles_with_keys } = CubDB . select ( db , min_key : { :articles , 0 } , max_key : { :articles , nil } ) This range selection works because our IDs are positive integers, and nil is greater than all numbers, so {:abc, nil} is greater than {:abc, 123} , but smaller than {:bcd, :123} .","ref":"howto.html#store-multiple-collections"},{"type":"extras","title":"How To - Save and restore a backup","doc":"Use the CubDB.back_up/2 function to create a backup of the current state of the database. Once the backup is completed, it can be opened by starting a new CubDB process using the target directory of the backup as its data directory: # Backup the current state of the database :ok = CubDB . back_up ( db , &quot;some/target/path&quot; ) # Open the backup as another CubDB process { :ok , copy } = CubDB . start_link ( data_dir : &quot;some/target/path&quot; )","ref":"howto.html#save-and-restore-a-backup"},{"type":"extras","title":"Upgrade Guide","doc":"The database format is completely backward compatible, so version v2 of CubDB can load databases created with v1 , and vice-versa. Upgrading from v1 to v2 requires a few code changes though, as some functions changed signature. When calling get_and_update , get_and_update_multi , and select the return value is not a {:ok, result} tuple anymore, but just result . The function get_and_update_multi does not take a fourth option argument anymore. The only available option was :timeout , which was now removed. In case you want to enforce a timeout for the update function, you can use a Task and Task.yield like explained here . The select function does not accept the :pipe and :reduce options anymore. Instead, it returns a lazy stream that can be used with functions in the Stream and Enum modules. For example: # This v1 code: { :ok , product } = CubDB . select ( db , [ min_key : :foo , max_key : :bar , pipe : [ map : fn { _ , val } -&gt; val end , filter : fn val -&gt; val &gt; 0 end ] , reduce : fn val , acc -&gt; val * acc end ] ) # Can be rewritten to this code in v2: product = CubDB . select ( db , min_key : :foo , max_key : :bar ) |&gt; Stream . map ( fn { _ , val } -&gt; val end ) |&gt; Stream . filter ( fn val -&gt; val &gt; 0 end ) |&gt; Enum . reduce ( fn val , acc -&gt; val * acc end ) The select function also does not support a :timeout option anymore, so it will be ignored if passed. In order to enforce a timeout, you can wrap the select in a Task and use Task.yield like shown here","ref":"upgrading.html"}]